/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the
 * distribution, and (3) all advertising materials mentioning features
 * or use of this software display the following acknowledgment:
 * ``This product includes software developed by the Politecnico di Torino
 * Telecomunication Networking Group.''
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Author:	Marco Mellia, Andrea Carpani
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 * The original TCPTRACE copyright can be found in tcptrace.copyright file.
 *
*/

#include <stdio.h>
#include "tstat.h"

extern unsigned long int fcount;
/*
*   histo e basta
*/

static struct histol *hl = NULL;

struct histo *
create_histo (char *nome, float min, float max, float delta)
{
  struct histo *thisto;
  int *tdati, num_col;
  struct histol *temphp;

  /* controllo sul delta corretto */

  num_col = ((max - min) / delta);
  /* printf("Intervallo [%f:%f] con %d sottointervalli di %f ciascuno\n", min, max, num_col, delta); */

  if (max != (min + (num_col * delta)))
    {
      printf ("Error: wrong 'delta' param when creating %s histogram \n", nome);
      return 0;
    }

  thisto = (struct histo *) MMmalloc (sizeof (struct histo), "create_histo");
  // memset(thisto,0,sizeof (struct histo));
  thisto->nome = nome;
  thisto->min = min;
  thisto->max = max;
  thisto->delta = delta;

  tdati = (int *) MMmalloc (sizeof (int) * (num_col + 2), "create_histo");
  // memset(tdati,0,(sizeof (int)*(num_col+2)));

  thisto->dati = tdati;

  /* aggiungo il puntatore a questa histo alla lista di tutte le histo */
  temphp =
    (struct histol *) MMmalloc (sizeof (struct histol), "create_histo");
  temphp->histop = thisto;
  temphp->next = hl;
  hl = temphp;

  return thisto;

}


struct histo *
create_histo2 (char *nome, float min, float max, int num_col)
{
  float delta;
  struct histo *thisto;

  delta = (max - min) / num_col;

  thisto = create_histo (nome, min, max, delta);

  if (!thisto)
    {
      return 0;
    }
  else
    {
      return thisto;
    }
}

int
clear_all_histo ()
{
  struct histol *temphp;

  temphp = hl;
  while (temphp != NULL)
    {
      clear_histo (temphp->histop);
      temphp = temphp->next;
    }
  return 1;
}

int
clear_histo (struct histo *phisto)
{
  int num_col, i;

  num_col = (((phisto->max) - (phisto->min)) / phisto->delta) + 2;
  for (i = 0; i < num_col; i++)
    phisto->dati[i] = 0;
  return 1;
}


int
print_all_histo ()
{
  struct histol *temphp;
  FILE *fp;
  char filename[200] = "";
  static unsigned long int last_fcount = 0;

  temphp = hl;
  while (temphp != NULL)
    {
      print_histo (temphp->histop, temphp->histop->nome, 1);
      temphp = temphp->next;
    }

  /* Faccio stampare anche i pacchetti ex-rumenta ora not_id_p */
  sprintf (filename, "%s/%s", curr_data_dir, "not_id_p");
  fp = fopen (filename, "w");
  fprintf (fp, "%lu\n", not_id_p);
  fclose (fp);
  not_id_p = 0;

  /* Printing the number of flow seen in the last time interval */
  sprintf (filename, "%s/%s", curr_data_dir, "flow_number");
  fp = fopen (filename, "w");
  fprintf (fp, "0 %lu\n", fcount-last_fcount);
  last_fcount = fcount;
  fclose (fp);
  return 1;
}

int
print_histo (struct histo *phisto, char *titolo, int flag)
{
  float min, max, delta;
  int num_row, i;
  struct stat fbuf;
  FILE *fp;
  char filename[200] = "";

  min = phisto->min;
  max = phisto->max;
  delta = phisto->delta;

  /* check directory status */
  if (stat (curr_data_dir, &fbuf) == -1)
    {
      printf ("Creating output dir %s\n", curr_data_dir);
      mkdir (curr_data_dir, 0775);
    }

  sprintf (filename, "%s/%s", curr_data_dir, titolo);
  fp = fopen (filename, "w");

  if (fp == NULL)
    {
      printf ("Could not open file %s\n", filename);
      return 0;
    }

  /* how many rows? +2 to include (premin e postmax) */
  num_row = ((max - min) / delta) + 2;

  /* Flag = 1 print start bin
     Flag = 0 if you do not want them, but rather the row number
   */

  for (i = 0; i < num_row; i++)
    {
      if (flag)
	{
	  fprintf (fp, "%.1f %d\n", (min + (i * delta) - delta),
		   phisto->dati[i]);
	}
      else
	{
	  fprintf (fp, "%d %d\n", i, phisto->dati[i]);
	}
    }
  fclose (fp);
  return 1;
}

int
print_text_histo (struct histo *phisto, char *titolo, char *labels)
{
  float min, max, delta;
  int num_col, i;
  char *llabels;


  min = phisto->min;
  max = phisto->max;
  delta = phisto->delta;
  llabels = labels;

  /* le righe da stampare sonu num_col +2 (premin e postmax) */
  num_col = ((max - min) / delta) + 2;

  printf ("\nResults for [%f:%f] with %d steps (%f)\n", min, max, num_col,
	  delta);
  printf ("*** %s ***\n", titolo);

  for (i = 0; i < num_col; i++)
    {
      /* stampo la riga */
      while ((*llabels != '!') && (*llabels != '\0'))
	{
	  printf ("%c", *llabels);
	  llabels++;
	}
      if (*llabels != '\0')
	{
	  llabels++;
	}
      else
	{
	  printf ("Val");
	}
      /* questo e' il valore */
      printf ("\t:%d\n", phisto->dati[i]);
    }
  return 1;
}

/*
*  dynamic histo 
*/

int
print_dhisto (struct dhisto *phisto, char *titolo, char *labels)
{
  struct dato *elem;
  float min, max, delta;
  int num_col, i;
  char *llabels;


  min = phisto->min;
  max = phisto->max;
  delta = phisto->delta;
  elem = phisto->dati;
  llabels = labels;

  /* le righe da stampare sonu num_col +2 (premin e postmax) */
  num_col = ((max - min) / delta) + 2;

  printf ("\nResults for [%f:%f] with %d steps (%f)\n", min, max, num_col,
	  delta);
  printf ("*** %s ***\n", titolo);
  for (i = 1; i <= num_col; i++)
    {
      while ((*llabels != '!') && (*llabels != '\0'))
	{
	  printf ("%c", *llabels);
	  llabels++;
	}
      if (*llabels != '\0')
	{
	  llabels++;
	}
      else
	{
	  printf ("Val");
	}
      printf (":\t%d\n", elem->val);
      elem = elem->next;
    }
  return 1;
}

int
add_dhisto (struct dhisto *phisto, float val)
{
  float min, max, delta;
  int num_col, i, j;
  struct dato *elem;

  min = phisto->min;
  max = phisto->max;
  delta = phisto->delta;
  elem = phisto->dati;

  num_col = (max - min) / delta;
  /* printf("Colonne: %d\n", num_col); */
  /* val < min lo metto e esco */
  if (val < min)
    {
      elem->val++;
/* 		printf("Adding %f to < %f \n", val, min); */
      return 1;
    }

  /* senno' incasella il valore da qualche parte */

  /* ...restano (((max-min)/delta)+1) colonne piu' "quella oltre il max" */
  for (i = 1; i <= num_col; i++)
    {
      if (val <= (min + (delta * i)))
	{
	  /* devo aumentare il contatore, per fare questo prima lo devo raggiungere */
	  elem = phisto->dati;
	  for (j = 1; j <= i; j++)
	    {
	      elem = elem->next;
	    }
/* 			printf("Adding %f to interval: %d\n", val, i); */
	  elem->val++;
	  return 1;
	}
    }

  /* se sono qua vuol dire che il dato e' oltre il max */
  elem = phisto->dati;

  /* con questo percorro pre-min, tutti gli intervalli e punto a 5post-max */
  elem = elem->next;
  for (j = 1; j < num_col; j++)
    {
      elem = elem->next;
    }
  elem = elem->next;

/* 	printf("Adding %f to > %f \n", val, max); */
  elem->val++;
  return 1;
}

struct dhisto *
create_dhisto2 (float min, float max, int num_col)
{
  float delta;
  struct dhisto *htemp;

  delta = (max - min) / num_col;

  htemp = create_dhisto (min, max, delta);

  if (!htemp)
    {
      return 0;
    }
  else
    {
      return htemp;
    }
}

struct dhisto *
create_dhisto (float min, float max, float delta)
{
  struct dhisto *htemp;
  struct dato *elem, *temp_dato;
  int i, num_col;

  htemp = (struct dhisto *) malloc (sizeof (struct dhisto));
  htemp->min = min;
  htemp->max = max;
  htemp->delta = delta;
  htemp->dati = NULL;

  elem = htemp->dati;

  /* controllo sul delta corretto */

  num_col = ((max - min) / delta);
  printf ("Intervallo [%f:%f] con %d sottointervalli di %f ciascuno\n", min,
	  max, num_col, delta);

  if (max != (min + (num_col * delta)))
    {
      printf ("Il delta fornito non va bene!\n");
      return 0;
    }


  /* aggiungo tutte le "celle dati" che mi servono
     * ce ne sono 2 per i dati "oltre ogni limite" (sotto min e sopra max)
     * e ((max-min)/delta)+1 all'interno dell'intervallone 
   */

  /* la prima la metto a mano. */
  temp_dato = (struct dato *) malloc (sizeof (struct dato));
  temp_dato->val = 0;
  temp_dato->next = NULL;
  elem = temp_dato;
  htemp->dati = temp_dato;


  /* ...restano (((max-min)/delta)+1) colonne piu' "quella oltre il max" */

  for (i = 0; i <= num_col; i++)
    {

      /* per prima cosa scivolo fino all'ultima cella */
      while (elem->next != NULL)
	{
	  elem = elem->next;
	}

      /* alla coda aggiungo un "dato" */
      temp_dato = (struct dato *) malloc (sizeof (struct dato));
      temp_dato->val = 0;
      temp_dato->next = NULL;
      elem->next = temp_dato;

    }

  return htemp;
}
