/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the
 * distribution, and (3) all advertising materials mentioning features
 * or use of this software display the following acknowledgment:
 * ``This product includes software developed by the Politecnico di Torino
 * Telecomunication Networking Group.''
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Author:	Marco Mellia, Andrea Carpani
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 * The original TCPTRACE copyright can be found in tcptrace.copyright file.
 *
*/


#include "tstat.h"


/* local routines */
double Average (double, int);
double Stdev (double, double, int);

/* to support some of the counters being long long on some platforms, use this */
/* macro... */
#ifdef HAVE_LONG_LONG
#define StatLineI(label,units,ul1,ul2)  \
(sizeof((ul1)) == SIZEOF_UNSIGNED_LONG_LONG_INT)?\
  StatLineI_LL((label),(units),(ul1),(ul2)):\
  StatLineI_L ((label),(units),(ul1),(ul2))
#else /* HAVE_LONG_LONG */
#define StatLineI StatLineI_L
#endif /* HAVE_LONG_LONG */



u_int
SynCount (tcp_pair * ptp)
{
  tcb *pab = &ptp->a2b;
  tcb *pba = &ptp->b2a;

  return (((pab->syn_count >= 1) ? 1 : 0) + ((pba->syn_count >= 1) ? 1 : 0));
}



u_int
FinCount (tcp_pair * ptp)
{
  tcb *pab = &ptp->a2b;
  tcb *pba = &ptp->b2a;

  return (((pab->fin_count >= 1) ? 1 : 0) + ((pba->fin_count >= 1) ? 1 : 0));
}



int
ConnComplete (tcp_pair * ptp)
{
  return (ptp->a2b.closed && ptp->b2a.closed);

/*     return(
		(SynCount(ptp) >= 2) &&
		(FinCount(ptp) >= 2) &&
        (
			((ptp->a2b.ack == ptp->b2a.seq) &&
			(ptp->b2a.ack == (ptp->a2b.seq+1)))
		||
			((ptp->a2b.ack == (ptp->b2a.seq+1)) &&
			(ptp->b2a.ack == ptp->a2b.seq))
		)
		);
 */
}


int
ConnReset (tcp_pair * ptp)
{
  return (ptp->a2b.reset_count + ptp->b2a.reset_count != 0);
}



double
Average (double sum, int count)
{
  return ((double) sum / ((double) count + .0001));
}



double
Stdev (double sum, double sum2, int n)
{
  double term;
  double term1;
  double term2;
  double retval;

  if (n <= 2)
    return (0.0);

  term1 = sum2;
  term2 = (sum * sum) / (double) n;
  term = term1 - term2;
  term /= (double) (n - 1);
  retval = sqrt (term);

/*     printf("Stdev(%f,%f,%d) is %f\n", sum,sum2,n,retval); */

  return (retval);
}

char *
elapsed2str (double etime)
{
  static char buf[80];
  u_long etime_secs;
  u_long etime_usecs;

  etime_secs = etime / 1000000.0;
  etime_usecs = 1000000 * (etime / 1000000.0 - (double) etime_secs);
  sprintf (buf, "%lu:%02lu:%02lu.%06lu",
	   etime_secs / (60 * 60),
	   etime_secs % (60 * 60) / 60,
	   (etime_secs % (60 * 60)) % 60, etime_usecs);
  return (buf);
}
