/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the
 * distribution, and (3) all advertising materials mentioning features
 * or use of this software display the following acknowledgment:
 * ``This product includes software developed by the Politecnico di Torino
 * Telecomunication Networking Group.''
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Author:	Marco Mellia, Andrea Carpani
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 * The original TCPTRACE copyright can be found in tcptrace.copyright file.
 *
*/


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <net/if.h>
#if __FreeBSD__ >= 2
#include <osreldate.h>
#if __FreeBSD_version >= 300000
#include <net/if_var.h>
#endif
#endif
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <ctype.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>

#ifdef linux
#ifdef strncpy
  /* stupid Linux (redhat?) bug in macro */
#undef strncpy
#endif /* strncpy */
#endif /* linux */

/* #include "memwatch.h" */
/* #include <mpatrol.h> */

/* IPv6 support */
#include "ipv6.h"

/* we want LONG LONG in some places */
#if SIZEOF_UNSIGNED_LONG_LONG_INT >= 8
#define HAVE_LONG_LONG
typedef unsigned long long int u_llong;
typedef long long int llong;
#else /* LONG LONG */
typedef unsigned long int u_llong;
typedef long int llong;
#endif /* LONG LONG */

/* lunghezza delle stringhe associate ai flussi */
#define STRMAXLEN 20

/* type for a TCP sequence number, ACK, FIN, or SYN */
typedef u_long seqnum;

/* length of a segment */
typedef u_long seglen;

/* type for a quadrant number */
typedef u_char quadnum;		/* 1,2,3,4 */

/* type for a TCP port number */
typedef u_short portnum;

/* type for an IP address */
/* IP address can be either IPv4 or IPv6 */
typedef struct ipaddr
{
  u_char addr_vers;		/* 4 or 6 */
  union
  {
    struct in_addr ip4;
#ifdef SUPPORT_IPV6
    struct in6_addr ip6;
#endif
  }
  un;
}
ipaddr;


/* type for a timestamp */
typedef struct timeval timeval;
#define ZERO_TIME(ptv)(((ptv)->tv_sec == 0) && ((ptv)->tv_usec == 0))
#define time2double(t) ((double) (t).tv_sec * 1000000 + (double) (t).tv_usec)

/* type for a Boolean */
typedef u_char Bool;
#define TRUE	1
#define FALSE	0
#define BOOL2STR(b) (b)?"TRUE":"FALSE"

/* ACK types */
enum t_ack
{ NORMAL = 1,			/* no retransmits, just advance */
  AMBIG = 2,			/* segment ACKed was rexmitted */
  CUMUL = 3,			/* doesn't advance */
  TRIPLE = 4,			/* triple dupack */
  NOSAMP = 5
};				/* covers retransmitted segs, no rtt sample */

/* type for an internal file pointer */
typedef struct mfile MFILE;

/* file name information */
extern char *cur_filename;	/* the current file */

extern unsigned long int fcount; /* total flow number number */

/* first and last packet timestamp */
extern timeval first_packet;
extern timeval last_packet;

/* counters */
extern u_long tcp_trace_count;
extern u_long udp_trace_count;


#ifdef OLD
/* test 2 IP addresses for equality */
#define IP_SAMEADDR(addr1,addr2) (((addr1).s_addr) == ((addr2).s_addr))

/* copy IP addresses */
#define IP_COPYADDR(toaddr,fromaddr) ((toaddr).s_addr = (fromaddr).s_addr)
#endif

typedef struct segment
{
  seqnum seq_firstbyte;		/* seqnumber of first byte */
  seqnum seq_lastbyte;		/* seqnumber of last byte */
  u_char retrans;		/* retransmit count */
  u_char acked;			/* how MANY times has has it been acked? */
  timeval time;			/* time the segment was sent */
  struct segment *next;
  struct segment *prev;
}
segment;

typedef struct quadrant
{
  segment *seglist_head;
  segment *seglist_tail;
  Bool full;
  struct quadrant *prev;
  struct quadrant *next;
}
quadrant;

typedef struct seqspace
{
  quadrant *pquad[4];
}
seqspace;

typedef struct tcb
{
  /* parent pointer */
  struct stcp_pair *ptp;
  struct tcb *ptwin;

  seqnum highest_seqno;		/* to identify the end of the data */
  Bool data_to_be_acked;	/* true if waiting for an ack */
  timeval payload_end_time;	/* time of last valid ack... */

  /* TCP information */
  seqnum ack;
  seqnum seq;
  seqnum syn;
  seqnum fin;
  seqnum fin_seqno;
  seqnum windowend;
  timeval time;

  /* TCP options */
  u_int mss;
  Bool f1323_ws;		/* did he request 1323 window scaling? */
  Bool f1323_ts;		/* did he request 1323 timestamps? */
  Bool fsack_req;		/* did he request SACKs? */
  u_char window_scale;

  /* statistics added */
  u_long data_bytes;
  u_long data_pkts;
  u_long data_pkts_push;
  u_long unique_bytes;         /* bytes sent (-FIN/SYN), excluding rexmits */
  u_int rexmit_bytes;
  u_int rexmit_pkts;
  u_long ack_pkts;
  u_long pureack_pkts;         /* mallman - pure acks, no data */
  u_int win_max;
  u_int win_min;
  u_int win_tot;
  u_int win_zero_ct;
  u_int min_seq;
  u_int max_seq;
  u_long packets;
  u_char syn_count;
  u_char fin_count;
  Bool closed;
  u_char reset_count;		/* resets SENT */
  u_int min_seg_size;
  u_int max_seg_size;
  u_int out_order_pkts;	/* out of order packets */
  u_int sacks_sent;		/* sacks returned */

  /* .a.c. rwnd stats */
  u_int rwnd_ini;
  u_int rwnd_tot;
  u_int rwnd_max;
  u_int rwnd_curr;
  u_int rwnd_curr_burst;

  /* .a.c. cwnd stats */
  u_int cwnd_max;
  u_int cwnd_flag;

  /* .a.c. burst stats */
  seqnum last_ooo_seq;
  u_long curr_ooo_burst;
  Bool curr_ooo_sf;
  u_long tot_ooob;		/* numero tot di burst ooo standard */
  u_int tot_ooob_sf;		/* numero tot di burst ooo sf */
  u_int tot_ooob_rwndz;	/* numero tot di burst ooo rwndz */
  seqnum last_dup_seq;
  u_long curr_dup_burst;
  Bool curr_dup_sf;
  u_long tot_dupb;		/* numero tot di burst dup standard */
  u_int tot_dupb_sf;	     /* numero tot di burst dup sf */
  u_int tot_dupb_rwndz;	     /* numero tot di burst dup rwndz */

  /* did I detect any "bad" tcp behavior? */
  /* at present, this means: */
  /*  - SYNs retransmitted with different sequence numbers */
  /*  - FINs retransmitted with different sequence numbers */
  Bool bad_behavior;

  /* added for initial window stats (for Mallman) */
  u_int initialwin_bytes;      /* initial window (in bytes) */
  u_int initialwin_segs;	    /* initial window (in segments) */
  Bool data_acked;		/* has any non-SYN data been acked? */

  /* added for (estimated) congestions window stats (for Mallman) */
  u_long cwin_max;
  u_long cwin_min;

  /* RTT stats for singly-transmitted segments */
  double rtt_last;		/* RTT as of last good ACK (microseconds) */
  u_long rtt_min;
  u_long rtt_max;
  double rtt_sum;		/* for averages */
  double rtt_sum2;		/* sum of squares, for stdev */
  u_int rtt_count;		/* for averages */
  /* RTT stats for multiply-transmitted segments */
  u_long rtt_min_last;
  u_long rtt_max_last;
  double rtt_sum_last;		/* from last transmission, for averages */
  double rtt_sum2_last;		/* sum of squares, for stdev */
  u_int rtt_count_last;	/* from last transmission, for averages */
  /* ACK Counters */
  u_long rtt_amback;	       /* ambiguous ACK */
  u_long rtt_cumack;	       /* segments only cumulativly ACKed */
  u_long rtt_nosample;         /* segments ACKED, but after retransmission */
  /* of earlier segments, so sample isn't */
  /* valid */
  u_long rtt_unkack;	       /* unknown ACKs  ??? */
  u_long rtt_dupack;	       /* duplicate ACKs */
  u_long rtt_triple_dupack;    /* triple duplicate ACKs */
  /* retransmission information */
  seqspace *ss;			/* the sequence space */
  u_long retr_max;		/* maximum retransmissions ct */
  u_long retr_min_tm;		/* minimum retransmissions time */
  u_long retr_max_tm;		/* maximum retransmissions time */
  double retr_tm_sum;		/* for averages */
  double retr_tm_sum2;		/* sum of squares, for stdev */
  u_long retr_tm_count;		/* for averages */

  /* Instantaneous throughput info */
  timeval thru_firsttime;	/* time of first packet this interval */
  u_long thru_bytes;		/* number of bytes this interval */
  u_long thru_pkts;		/* number of packets this interval */
  timeval thru_lasttime;	/* time of previous segment */

  /* data transfer time stamps - mallman */
  timeval first_data_time;
  timeval last_data_time;


  /* for tracking unidirectional idle time */
  timeval last_time;		/* last packet SENT from this side */
  u_llong idle_max;		/* maximum idle time observed (usecs) */

}
tcb;


typedef u_long hash;

typedef struct
{
  ipaddr a_address;
  ipaddr b_address;
  portnum a_port;
  portnum b_port;
  hash hash;
}
tcp_pair_addrblock;


struct stcp_pair
{

  /* endpoint identification */
  tcp_pair_addrblock addr_pair;

  /* connection naming information */
  Bool internal_src;
  Bool internal_dst;

  /* connection information */
  timeval first_time;
  timeval last_time;
  u_long packets;
  tcb a2b;
  tcb b2a;
};
typedef struct stcp_pair tcp_pair;

typedef struct tcphdr tcphdr;

typedef struct ptp_snap
{
  tcp_pair_addrblock addr_pair;	/* just a copy */
  struct ptp_snap *next;
  tcp_pair *ptp;
  tcp_pair **ttp_ptr;
}
ptp_snap;

extern int num_tcp_pairs;	/* how many pairs are in use */
extern tcp_pair **ttp;		/* array of pointers to allocated pairs */

/* minimal support for UDP "connections" */
typedef struct ucb
{
  /* parent pointer */
  struct sudp_pair *pup;
  struct ucb *ptwin;

  /* statistics added */
  u_llong data_bytes;
  u_llong packets;
  u_long min_dg_size;
  u_long max_dg_size;
}
ucb;


typedef tcp_pair_addrblock udp_pair_addrblock;
struct sudp_pair
{
  /* endpoint identification */
  udp_pair_addrblock addr_pair;

  /* connection information */
  timeval first_time;
  timeval last_time;
  u_llong packets;
  ucb a2b;
  ucb b2a;

  /* linked list of usage */
  struct sudp_pair *next;
};
typedef struct sudp_pair udp_pair;
typedef struct udphdr udphdr;


extern int num_udp_pairs;	/* how many pairs are in use */
extern udp_pair **utp;		/* array of pointers to allocated pairs */


/* option flags */
extern Bool internal_src;
extern Bool internal_dst;

extern Bool warn_ooo;
extern Bool warn_IN_OUT;
extern Bool warn_printtrunc;
extern Bool warn_printbadmbz;
extern Bool warn_printbad_syn_fin_seq;
extern Bool save_tcp_data;
extern Bool do_udp;
extern int debug;
extern u_long pnum;

extern u_long ctrunc;
extern timeval current_time;

/* first and last packet timestamp */
extern timeval first_packet;
extern timeval last_packet;


#define MAX_NAME 20

/* external routine decls */
double sqrt (double x);
char *ether_ntoa (struct ether_addr *e);
void free (void *);
int finite (double);

/* global routine decls */
void QuitSig (int);
void Usr1Sig (int);
void *MallocZ (int);
void *ReallocZ (void *oldptr, int obytes, int nbytes);
void trace_init (void);
void trace_done (void);
void seglist_init (tcb *);
tcp_pair *dotrace_2 (struct ip *, struct tcphdr *ptcp, void *plast);
double elapsed (timeval, timeval);
void tv_sub (struct timeval *plhs, struct timeval rhs);
void tv_add (struct timeval *plhs, struct timeval rhs);
int tv_cmp (struct timeval lhs, struct timeval rhs);
Bool tv_same (struct timeval lhs, struct timeval rhs);
char *elapsed2str (double etime);
int ConnReset (tcp_pair *);
int ConnComplete (tcp_pair *);
double Average (double sum, int count);
double Stdev (double sum, double sum2, int n);
u_int SynCount (tcp_pair * ptp);
u_int FinCount (tcp_pair * ptp);
char *ts2ascii (timeval *);
char *ts2ascii_date (timeval *);
char *ServiceName (portnum);
char *HostName (ipaddr);
char *HostAddr (ipaddr);
char *HostLetter (u_int);
char *NextHostLetter (void);
char *EndpointName (ipaddr, portnum);
int rexmit (tcb *, seqnum, seglen, Bool *);
enum t_ack ack_in (tcb *, seqnum, unsigned tcp_data_length);
void DoThru (tcb * ptcb, int nbytes);
struct mfile *Mfopen (char *fname, char *mode);
void Minit (void);
int Mfileno (MFILE * pmf);
int Mvfprintf (MFILE * pmf, char *format, va_list ap);
int Mfwrite (char *buf, u_long size, u_long nitems, MFILE * pmf);
long Mftell (MFILE * pmf);
int Mfseek (MFILE * pmf, long offset, int ptrname);
int Mfprintf (MFILE * pmf, char *format, ...);
int Mfflush (MFILE * pmf);
int Mfclose (MFILE * pmf);
int Mfpipe (int pipes[]);
struct tcp_options *ParseOptions (struct tcphdr *ptcp, void *plast);
FILE *CompOpenHeader (char *filename);
FILE *CompOpenFile (char *filename);
void CompCloseFile (char *filename);
void CompFormats (void);
int CompIsCompressed (void);
Bool FileIsStdin (char *filename);
struct tcb *ptp2ptcb (tcp_pair * ptp, struct ip *pip, struct tcphdr *ptcp);
void IP_COPYADDR (ipaddr * toaddr, ipaddr fromaddr);
int IP_SAMEADDR (ipaddr addr1, ipaddr addr2);
void PcapSavePacket (char *filename, struct ip *pip, void *plast);
void StringToArgv (char *buf, int *pargc, char ***pargv);
void CopyAddr (tcp_pair_addrblock *, struct ip *pip, portnum, portnum);
int WhichDir (tcp_pair_addrblock *, tcp_pair_addrblock *);
int SameConn (tcp_pair_addrblock *, tcp_pair_addrblock *, int *);
Bool ip_cksum_valid (struct ip *pip, void *plast);
Bool tcp_cksum_valid (struct ip *pip, struct tcphdr *ptcp, void *plast);
Bool udp_cksum_valid (struct ip *pip, struct udphdr *pudp, void *plast);
ipaddr *str2ipaddr (char *str);
int IPcmp (ipaddr * pipA, ipaddr * pipB);


/* routines added by .a.c. */
void make_conn_stats (tcp_pair * ptp_save, Bool flusso_nc);
void trace_done2 (void);

/* UDP support routines */
void udptrace_init (void);
void udptrace_done (void);
udp_pair *udpdotrace (struct ip *pip, struct udphdr *pudp, void *plast);

/* TCP flags macros */
#define SYN_SET(ptcp)((ptcp)->th_flags & TH_SYN)
#define FIN_SET(ptcp)((ptcp)->th_flags & TH_FIN)
#define ACK_SET(ptcp)((ptcp)->th_flags & TH_ACK)
#define RESET_SET(ptcp)((ptcp)->th_flags & TH_RST)
#define PUSH_SET(ptcp)((ptcp)->th_flags & TH_PUSH)
#define URGENT_SET(ptcp)((ptcp)->th_flags & TH_URG)
#define FLAG6_SET(ptcp)((ptcp)->th_flags & 0x40)
#define FLAG7_SET(ptcp)((ptcp)->th_flags & 0x80)
#define CWR_SET(ptcp)((ptcp)->th_x2 & TH_CWR)
#define ECN_ECHO_SET(ptcp)((ptcp)->th_x2 & TH_ECN_ECHO)

/* Vuole la notazione numeri e punti */
#define INTERNAL_IP(ip) ((inet_addr(ip)  & inet_addr("255.255.0.0")) == inet_addr(internal_net))
/* Vuole la notazione "binaria" */
#define INTERNAL_IP_BIN(ip) ((ip & inet_addr("255.255.0.0")) == inet_addr(internal_net))

/* connection directions */
#define A2B 1
#define B2A -1


/*macros for maintaining the seqspace used for rexmit*/
#define QUADSIZE	(0x40000000)
#define QUADNUM(seq)	((seq>>30)+1)
#define IN_Q1(seq)	(QUADNUM(seq)==1)
#define IN_Q2(seq)	(QUADNUM(seq)==2)
#define IN_Q3(seq)	(QUADNUM(seq)==3)
#define IN_Q4(seq)	(QUADNUM(seq)==4)
#define FIRST_SEQ(quadnum)	(QUADSIZE*(quadnum-1))
#define LAST_SEQ(quadnum)	((QUADSIZE-1)*quadnum)
#define BOUNDARY(beg,fin) (QUADNUM((beg)) != QUADNUM((fin)))


/* physical layers currently understood					*/
#define PHYS_ETHER	1
#define PHYS_FDDI       2

/*
 * SEQCMP - sequence space comparator
 *	This handles sequence space wrap-around. Overlow/Underflow makes
 * the result below correct ( -, 0, + ) for any a, b in the sequence
 * space. Results:	result	implies
 *			  - 	 a < b
 *			  0 	 a = b
 *			  + 	 a > b
 */
#define	SEQCMP(a, b)		((long)(a) - (long)(b))
#define	SEQ_LESSTHAN(a, b)	(SEQCMP(a,b) < 0)
#define	SEQ_GREATERTHAN(a, b)	(SEQCMP(a,b) > 0)


/* SACK TCP options (not an RFC yet, mostly from draft and RFC 1072) */
/* I'm assuming, for now, that the draft version is correct */
/* sdo -- Tue Aug 20, 1996 */
#define	TCPOPT_SACK_PERM 4	/* sack-permitted option */
#define	TCPOPT_SACK      5	/* sack attached option */
#define	MAX_SACKS       10	/* max number of sacks per segment (rfc1072) */
typedef struct sack_block
{
  seqnum sack_left;		/* left edge */
  seqnum sack_right;		/* right edge */
}
sack_block;

#define MAX_UNKNOWN 16
typedef struct opt_unknown
{
  u_char unkn_opt;
  u_char unkn_len;
}
opt_unknown;

/* RFC 1323 TCP options (not usually in tcp.h yet) */
#define	TCPOPT_WS	3	/* window scaling */
#define	TCPOPT_TS	8	/* timestamp */

/* other options... */
#define	TCPOPT_ECHO		6	/* echo (rfc1072) */
#define	TCPOPT_ECHOREPLY	7	/* echo (rfc1072) */
#define TCPOPT_TIMESTAMP	8	/* timestamps (rfc1323) */
#define TCPOPT_CC		11	/* T/TCP CC options (rfc1644) */
#define TCPOPT_CCNEW		12	/* T/TCP CC options (rfc1644) */
#define TCPOPT_CCECHO		13	/* T/TCP CC options (rfc1644) */

/* RFC 2481 (ECN) IP and TCP flags (not usually defined yet) */
#define IPTOS_ECT	0x02	/* ECN-Capable Transport */
#define IPTOS_CE	0x01	/* Experienced Congestion */

#define TH_ECN_ECHO	0x02	/* Used by receiver to echo CE bit */
#define TH_CWR		0x01	/* Congestion Window Reduced */



/* some compilers seem to want to make "char" unsigned by default, */
/* which is breaking stuff.  Rather than introduce (more) ugly */
/* machine dependencies, I'm going to FORCE some chars to be */
/* signed... */
typedef signed char s_char;

struct tcp_options
{
  short mss;			/* maximum segment size         */
  s_char ws;			/* window scale (1323)          */
  long tsval;			/* Time Stamp Val (1323)        */
  long tsecr;			/* Time Stamp Echo Reply (1323) */

  Bool sack_req;		/* sacks requested              */
  s_char sack_count;		/* sack count in this packet */
  sack_block sacks[MAX_SACKS];	/* sack blocks */

  /* echo request and reply */
  /* assume that value of -1 means unused  (?) */
  u_long echo_req;
  u_long echo_repl;

  /* T/TCP stuff */
  /* assume that value of -1 means unused  (?) */
  u_long cc;
  u_long ccnew;
  u_long ccecho;

  /* record the stuff we don't understand, too */
  char unknown_count;		/* number of unknown options */
  opt_unknown unknowns[MAX_UNKNOWN];	/* unknown options */
};


/* packet-reading options... */
/* the type for a packet reading routine */
typedef int pread_f (struct timeval *, int *, int *, void **,
		     int *, struct ip **, void **);

/* give the prototypes for the is_GLORP() routines supported */
#ifdef GROK_SNOOP
pread_f *is_snoop (void);
#endif /* GROK_SNOOP */
#ifdef GROK_NETM
pread_f *is_netm (void);
#endif /* GROK_NETM */
#ifdef GROK_TCPDUMP
pread_f *is_tcpdump (void);
#endif /* GROK_TCPDUMP */
#ifdef GROK_ETHERPEEK
pread_f *is_EP (void);
#endif /* GROK_ETHERPEEK */
#ifdef GROK_NS
pread_f *is_ns (void);
#endif /* GROK_NS */
#ifdef GROK_NLANR
pread_f *is_nlanr (void);
#endif /* GROK_NLANR */
#ifdef GROK_DAG
pread_f *is_dag (void);
int
pread_dag (struct timeval *ptime,
	   int *plen,
	   int *ptlen,
	   void **pphys, int *pphystype, struct ip **ppip, void **pplast);
#endif /* GROK_DAG */


/* I've had problems with the memcpy function that gcc stuffs into the program
   and alignment problems.  This should fix it! */
void *MemCpy (void *p1, void *p2, size_t n);	/* in tcptrace.c */
#define memcpy(p1,p2,n) MemCpy(p1,p2,n);


/*
 * timeval compare macros
 */
#define tv_ge(lhs,rhs) (tv_cmp((lhs),(rhs)) >= 0)
#define tv_gt(lhs,rhs) (tv_cmp((lhs),(rhs)) >  0)
#define tv_le(lhs,rhs) (tv_cmp((lhs),(rhs)) <= 0)
#define tv_lt(lhs,rhs) (tv_cmp((lhs),(rhs)) <  0)
#define tv_eq(lhs,rhs) (tv_cmp((lhs),(rhs)) == 0)

/* handy constants */
#define US_PER_SEC 1000000	/* microseconds per second */
#define MS_PER_SEC 1000		/* milliseconds per second */


/*
 * Macros to simplify access to IPv4/IPv6 header fields
 */
#define PIP_VERS(pip) (((struct ip *)(pip))->ip_v)
#ifdef SUPPORT_IPV6
   #define PIP_ISV6(pip) (PIP_VERS(pip) == 6)
#else
   #define PIP_ISV6(pip) FALSE
#endif
#define PIP_ISV4(pip) (PIP_VERS(pip) == 4)
#define PIP_V6(pip) ((struct ipv6 *)(pip))
#define PIP_V4(pip) ((struct ip *)(pip))
#define PIP_EITHERFIELD(pip,fld4,fld6) \
   (PIP_ISV4(pip)?(PIP_V4(pip)->fld4):(PIP_V6(pip)->fld6))
#define PIP_LEN(pip) (PIP_EITHERFIELD(pip,ip_len,ip6_lngth))

/*
 * Macros to simplify access to IPv4/IPv6 addresses
 */
#define ADDR_VERSION(paddr) ((paddr)->addr_vers)
#define ADDR_ISV4(paddr) (ADDR_VERSION((paddr)) == 4)
#ifdef SUPPORT_IPV6
  #define ADDR_ISV6(paddr) (ADDR_VERSION((paddr)) == 6)
#else
  #define ADDR_ISV6(paddr) (FALSE)
#endif
struct ipaddr *IPV4ADDR2ADDR (struct in_addr *addr4);
struct ipaddr *IPV6ADDR2ADDR (struct in6_addr *addr6);

/*
 * Macros to check for congestion experienced bits
 */
#define IP_CE(pip) (((struct ip *)(pip))->ip_tos & IPTOS_CE)
#define IP_ECT(pip) (((struct ip *)(pip))->ip_tos & IPTOS_ECT)

/*
 * fixes for various systems that aren't exactly like Solaris
 */
#ifndef IP_MAXPACKET
#define IP_MAXPACKET 65535
#endif /* IP_MAXPACKET */

#ifndef ETHERTYPE_REVARP
#define ETHERTYPE_REVARP        0x8035
#endif /* ETHERTYPE_REVARP */

/* added by .a.c. */
/* ac */

#define MAX_ADX_SLOTS 70001 /* hash size for the ip addresses hit counter */
#define ADDR_MASK 0x00ffffff /* mask to count IP addresses - inet order*/


#define MAX_INTERNAL_HOSTS 20

/* IDLE_TIME in microsecondi */
/* 1 min */
/* #define IDLE_TIME 60000000 */
/* 10 min */
#define IDLE_TIME 600000000

/* maximum number of concurrent connection sotred in the vector TTP */
#define MAX_TCP_PAIRS 500001

/* connection records are stored in a hash table.  Buckets are linked	*/
/* lists sorted by most recent access.					*/
#ifdef SMALL_TABLE
#define HASH_TABLE_SIZE 1021	/* oughta be prime */
#else /* SMALL_TABLE */
#define HASH_TABLE_SIZE 200007	/* oughta be prime */
#endif /* SMALL_TABLE */
/* dept of the linear serch in the hash... */
#define LIST_SEARCH_DEPT 50

/* STEP per la creazione delle histo in microsecondi */
/* 300000000 = 5 min */
#define MAX_TIME_STEP 300000000.0
/* 900000000 = 15 min */
#define MAX_TIME_STEP 900000000.0

#include "histo.h"

#ifndef add_histo
inline static int add_histo (struct histo *p, float val);
inline static int
add_histo (struct histo *p, float val)
{

  int i;

  /* val < min lo metto e esco */
  if (val < p->min)
    {
      p->dati[0]++;
      /* printf("Adding %f to < %f \n", val, min); */
      return 1;
    }


  /* con questo sporco trucco includo anche il limite max */
  if (val >= p->max)
    {
      /* printf("Adding %f to interval: %d\n", val, (i-1)); */
      p->dati[(int) ((p->max - p->min) / p->delta + 1)]++;
      return 1;
    }

  i = ((int) val + p->delta - p->min) / (int) p->delta;
  /* printf("Adding %f to interval: %d\n", val, (i)); */
  p->dati[i]++;

  return 1;
}
#endif

char *get_basename (char *filename);
char *basename;
char curr_data_dir[200];
unsigned int step;

struct histo *test;
struct histo *protocol_ent;
struct histo *protocol_usc;
struct histo *tcp_port_src_ent;
struct histo *tcp_port_src_usc;
struct histo *tcp_port_dst_ent;
struct histo *tcp_port_dst_usc ;

struct histo *tcp_port_syn_src_ent;
struct histo *tcp_port_syn_src_usc;
struct histo *tcp_port_syn_dst_ent;
struct histo *tcp_port_syn_dst_usc ;

struct histo *udp_port_dst_ent;
struct histo *udp_port_dst_usc;

struct histo *tcp_opts;

struct histo *tcp_mss_a;
struct histo *tcp_mss_b;
struct histo *tcp_mss_used;

struct histo *tcp_rwnd_ini;
struct histo *tcp_rwnd_avg;
struct histo *tcp_rwnd_max;

struct histo *tcp_cwnd;

struct histo *tcp_cl_b_l_usc;
struct histo *tcp_cl_b_l_ent;
struct histo *tcp_cl_b_s_usc;
struct histo *tcp_cl_b_s_ent;
struct histo *tcp_cl_p_usc;
struct histo *tcp_cl_p_ent;

struct histo *tcp_cl_b_l_a2b;
struct histo *tcp_cl_b_l_b2a;
struct histo *tcp_cl_b_s_a2b;
struct histo *tcp_cl_b_s_b2a;
struct histo *tcp_cl_p_a2b;
struct histo *tcp_cl_p_b2a;

struct histo *ip_len_usc;
struct histo *ip_len_ent;
struct histo *ip_ttl_usc;
struct histo *ip_ttl_ent;
struct histo *ip_tos_usc;
struct histo *ip_tos_ent;
struct histo *tcp_cl_ent;
struct histo *tcp_cl_usc;

struct histo *rtt_min_usc;
struct histo *rtt_min_ent;
struct histo *rtt_min_a2b;
struct histo *rtt_min_b2a;

struct histo *rtt_avg_usc;
struct histo *rtt_avg_ent;
struct histo *rtt_avg_a2b;
struct histo *rtt_avg_b2a;

struct histo *rtt_max_usc;
struct histo *rtt_max_ent;
struct histo *rtt_max_a2b;
struct histo *rtt_max_b2a;

struct histo *rtt_stdev_usc;
struct histo *rtt_stdev_ent;
struct histo *rtt_stdev_a2b;
struct histo *rtt_stdev_b2a;

struct histo *rtt_cnt_usc;
struct histo *rtt_cnt_ent;
struct histo *rtt_cnt_a2b;
struct histo *rtt_cnt_b2a;

struct histo *ooob_a2b;
struct histo *ooob_b2a;
struct histo *ooob_usc;
struct histo *ooob_ent;
struct histo *ooob_zm_usc;
struct histo *ooob_zm_ent;

struct histo *dupb_a2b;
struct histo *dupb_b2a;
struct histo *dupb_usc;
struct histo *dupb_ent;

struct histo *ooob_sf_a2b;
struct histo *ooob_sf_b2a;
struct histo *ooob_sf_usc;
struct histo *ooob_sf_ent;

struct histo *dupb_sf_a2b;
struct histo *dupb_sf_b2a;
struct histo *dupb_sf_usc;
struct histo *dupb_sf_ent;

struct histo *ooob_rwndz_a2b;
struct histo *ooob_rwndz_b2a;
struct histo *ooob_rwndz_usc;
struct histo *ooob_rwndz_ent;

struct histo *dupb_rwndz_a2b;
struct histo *dupb_rwndz_b2a;
struct histo *dupb_rwndz_usc;
struct histo *dupb_rwndz_ent;

struct histo *tot_time;
timeval last_time_step;
timeval last_cleaned;
unsigned long not_id_p;

void ooo_stat_add (struct ip *pip, tcb * thisdir, int dir);
void dup_stat_add (struct ip *pip, tcb * thisdir, int dir);
void ooo_stat_sf_add (struct ip *pip, tcb * thisdir, int dir);
void dup_stat_sf_add (struct ip *pip, tcb * thisdir, int dir);
void ooo_stat_rwndz_add (struct ip *pip, tcb * thisdir, int dir);
void dup_stat_rwndz_add (struct ip *pip, tcb * thisdir, int dir);
int LoadInternalNets (char *file);
void free_tp (tcp_pair * ptp_save);
Bool internal_ip (struct in_addr adx);
Bool internal_ip_string (char *adx);
unsigned long tot_conn;
void trace_done_periodic ();
void create_new_outfiles (char *filename);
void freequad (quadrant ** ppquad);
void acstrncpy (char *d, char *s, int n);
void DumpCore ();
int UsingFreedPtpsnap (ptp_snap * ptph);
void dump_internal_stats (int err);
